/******************************************************************************
 *
 * dirhive.h - Dump NT registry hives keys. Part of the DirTools project.
 *
 * $Revision: 1.3 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifndef _DIRHIVE_H_
#define _DIRHIVE_H_ 1

#include <sys/types.h>
#include <stdint.h>

#define KEY_ROOT   0x2c         /* Type ID of ROOT key node */
#define KEY_NORMAL 0x20         /* Normal nk key */

#define INVALID_HEADER_BLOCK 1
#define UNUSED_HEADER_BLOCK 2
#define UNKNOWN_HEADER_BLOCK 4

#define HGLOBAL_OFFSET 0x1000
#define HBIN_HEADER_SIZE 0x20
#define HALLOC_SIZE 100000

#define HIVE_NAME_LENGTH (0x1fc - 0x2c)

/* Structures definitions and comments taken/adapted from Peter Nordahl
 * chntpw "Change NT password" utility
 * http://home.eunet.no/~pnordahl/ntpasswd/
 */


/* The first page of the registry file is some kind of header, lot of
 * it's contents is unknown, and seems to be mostly NULLs anyway.
 * Note also, that this is the only place in the registry I've been
 * able to find _any_ kind of checksumming
 */

struct regf_header
{

     uint32_t id;               /* 0x00000000 D-Word  ID:
                                 *  ASCII-"regf" = 0x66676572 */
     uint32_t unknown1;         /* 0x00000004 D-Word  ???? */
     uint32_t unknown2;         /* 0x00000008 D-Word  ????
                                 * Always the same value as at 0x00000004  */
     uint64_t timestamp;        /* 0x0000000C    Q-Word 
                                 * last modify date in WinNT date-format */
     uint32_t unknown3;         /* 0x00000014 D-Word  1 */
     uint32_t unknown4;         /* 0x00000018 D-Word  3 -
                                 * probably version #. 2 in NT3.51 */
     uint32_t unknown5;         /* 0x0000001C D-Word  0 */
     uint32_t unknown6;         /* 0x00000020 D-Word  1 */
     uint32_t ofs_rootkey;      /* 0x00000024 D-Word
                                 * Offset of 1st key record */
     uint32_t filesize;         /* 0x00000028 D-Word  Size of
                                 * the data-blocks (Filesize-4kb) */
     uint32_t unknown7;         /* 0x0000002C D-Word  1 */
     char name[HIVE_NAME_LENGTH]; /* Seems like the hive's name is
                                 * buried here, max len unknown */
     uint32_t checksum;         /* 0x000001FC D-Word  Sum of all
                                 * D-Words from 0x00000000 to 0x000001FB */
} __attribute__ ((packed));


/* The page header, I don't know if the 14 "dummy" bytes has a meaning,
 * they seem to be mostly NULLS
 */

struct hbin_page
{

     uint32_t id;               /* 0x0000       D-Word  ID:
                                 * ASCII-"hbin" = 0x6E696268  */
     uint32_t ofs_from1;        /* 0x0004       D-Word
                                 * Offset from the 1st hbin-Block */
     uint32_t ofs_next;         /* 0x0008       D-Word Offset to the
                                 * next hbin-Block (from THIS ONE)  */
     char dummy1[14];
     uint32_t len_page;         /* 0x001C       D-Word  Block-size???
                                 * Don't look like it, I only use the
                                 * next-offset in this program  */
     char data[1];              /* 0x0020   First data block starts here */

} __attribute__ ((packed));


/* Security descriptor. I know how it's linked, but don't know
 * how the real security data is constructed, it may as well
 * be like the higher level security structs defined by MS in its
 * includes & NT docs. Currently, I have no use for it.
 * Note that keys sharing the exact same security settings will
 * most likely point to the same security descriptor, thus
 * saving space and making it fast to make objects inherit settings
 * (is inheritance supported? they speak of security inheritance as a "new"
 * feature in the filesystem on NT5, even though I think it was
 *  also supported by the lower levels in the earlier versions)
 */

struct sk_key
{

     uint16_t id;               /* 0x0000       Word    ID:
                                 * ASCII-"sk" = 0x6B73        */
     uint16_t dummy1;           /* 0x0002       Word    Unused     */
     uint32_t ofs_prevsk;       /* 0x0004      D-Word  Offset of 
                                 * previous "sk"-Record */
     uint32_t ofs_nextsk;       /* 0x0008      D-Word
                                 * Offset of next "sk"-Record     */
     uint32_t no_usage;         /* 0x000C      D-Word  usage-counter     */
     uint32_t len_sk;           /* 0x0010      D-Word
                                 * Size of "sk"-record in bytes   */
     char data[1];              /* Security data up to len_sk bytes   */

} __attribute__ ((packed));


/* This is the key node (ie directory) descriptor, can contain subkeys
 * and/or values.
 * Note that for values, the count is stored here, but for subkeys
 * there's a count both here and in the offset-table (lf or li struct).
 * What happens if these mismatch is not known.
 * What's the classname thingy? Can't remember seeing that used in
 * anything I've looked at.
 */
struct nk_key
{

     /* Offset       Size    Contents */
     uint16_t id;               /*  0x0000   Word    ID:
                                 * ASCII-"nk" = 0x6B6E                */
     uint16_t type;             /*  0x0002   Word
                                 * for the root-key: 0x2C, otherwise 0x20 */
     uint64_t timestamp;        /*  0x0004 Q-Word  write-date/time
                                 * in windows nt notation */
     char dummy0[4];
     uint32_t ofs_parent;       /*  0x0010  D-Word
                                 * Offset of Owner/Parent key             */
     uint32_t no_subkeys;       /*  0x0014  D-Word  number of sub-Keys  */
     char dummy1[4];
     uint32_t ofs_lf;           /*  0x001C  D-Word  Offset of the
                                 * sub-key lf-Records       */
     char dummy2[4];
     uint32_t no_values;        /*  0x0024  D-Word  number of values */
     uint32_t ofs_vallist;      /*  0x0028  D-Word  Offset of the Value-List */
     uint32_t ofs_sk;           /*  0x002C  D-Word  Offset of the sk-Record */
     uint32_t ofs_classnam;     /*  0x0030  D-Word  Offset of the Class-Name */
     char dummy3[16];
     uint32_t dummy4;           /*  0x0044  D-Word  Unused (data-trash) */
     uint16_t len_name;         /*  0x0048   Word    name-length */
     uint16_t len_classnam;     /*  0x004A   Word    class-name length */
     char keyname[1];           /*  0x004C      ????    key-name */
} __attribute__ ((packed));


/* Hive definition, allocated by openHive(), dealloc by closeHive()
 * contains state data, must be passed in all functions
 */

struct hive
{
     char *filename;            /* Hives filename */
     int dump_any;
     int guess_root_path;
     wchar_t hive_name[HIVE_NAME_LENGTH + 1];
     char *root_path;
     int filedesc;              /* File descriptor
                                 * (only valid if state == OPEN) */
     off_t size;                /* Hives size (filesize) in bytes */
     off_t h1offset;
     unsigned char *buffer;     /* Files raw contents */
};

#endif /* _DIRHIVE_H_ */
