/******************************************************************************
 *
 * dirmft.c - Dump NTFS files informations. Part of the DirTools project.
 *
 * $Revision: 1.3 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#ifdef _WIN32
#include <windows.h>
#else /* _WIN32 */
#include "wincompat.h"
#endif /* _WIN32 */

#include "dt_error.h"
#include "structure.h"
#include "diskio.h"
#include "output.h"
#include "info.h"
#include "utility.h"
#include "macros.h"
#include "sysheader.h"
#include "sd_dump.h"
#include "dt_locale.h"
#ifdef _WIN32
#include "winsid.h"
#endif /* _WIN32 */

#ifndef _fcloseall
int _fcloseall (void);
#endif

#define COMPACT_HEADER "MFT Index|FileName|MD5|# Links|Size|Attributes|\
Creation|Last Write|Attribute Modification|Last Access|# Data Streams|\
Owner|Group|# DACE|# SACE|[SID DACE|Rights|...][SID SACE|Audits|...]\
[Data Stream Name|Size|...]\n"

/**********************************************************************
 *
 *
 *
 *
 *
 *
 *********************************************************************/
void
Usage (char *pProgName)
{
#ifdef _WIN32
     fprintf (stderr,
              "Usage: %s [-v][-m][-h][-e][-n][-i MFT_index][-u user_file]"
              "[-D][-d debug_file][-x error_file][-o output_file][-8]"
              " drive_letter\n",
              pProgName);
#else /* _WIN32 */
     fprintf (stderr, "Usage: %s [-v][-m][-h][-e][-n][-i MFT_index]"
              "[-D][-d [debug_file]][-x error_file][-o output_file][-8]"
              "[-s skip] file\n",
              pProgName);
#endif /* _WIN32 */
}

/**********************************************************************
 *
 *
 *
 *
 *
 *
 *********************************************************************/
void
DisplayHelp (char *pProgName)
{
     fprintf (stderr, "%s: Dump NTFS files informations\n\n", pProgName);
     Usage (pProgName);
     fprintf (stderr, "\nPackage version: %s\n", PACKAGE_STRING);
     fprintf (stderr, "\n-h : help\n");
     fprintf (stderr, "-v : verbose mode\n");
     fprintf (stderr,
              "-m : process md5 hash of the data streams "
              "(regular and alternates)\n");
     fprintf (stderr, "-e : process only erased files\n");
     fprintf (stderr, "-n : do not check parent directory sequence number\n");
     fprintf (stderr, "-i : output a single record given its index\n");
     fprintf (stderr, "     this option override the -e option\n");
#ifdef _WIN32
     fprintf (stderr,
              "-u : create a table to convert in human readable format SID\n");
#endif /* _WIN32 */
     fprintf (stderr, "-d, -D : debug mode, -D dump each MFT record read\n");
#ifndef _WIN32
     fprintf (stderr, "         default output is sdterr\n");
#endif
     fprintf (stderr, "-x : output error to file instead of stderr\n");
     fprintf (stderr, "-o : results sent to file instead of stdout\n");
#ifndef _WIN32
     fprintf (stderr, "-s : skip bytes when reading file\n");
#endif /* _WIN32 */
     fprintf (stderr, "-8 : output non-ascii characters using UTF-8\n");
     fprintf (stderr, "\nOutput format :\n");
     fprintf (stderr, COMPACT_HEADER);
     fprintf (stderr, "\nAll times are UTC according to local computer.\n");
     fprintf (stderr, "\n  File Attributes :\n");
     fprintf (stderr,
              "    Ro : Read-Only, Hi : Hidden, Sy : System, "
              "Di : Directory, Ar : Archive,\n");
     fprintf (stderr,
              "    En : Encrypted, No : Normal, Te : Temporary, "
              "Sp : Sparse file,\n");
     fprintf (stderr,
              "    Re : Reparse point, Co : Compressed, Of : Offline, "
              "Ni : Not content indexed\n");
     sd_help (stderr);
}

/**********************************************************************
 *
 *
 *
 *
 *
 *
 *********************************************************************/
     int
main (int argc, char **argv)
{
     int iArg, iOptInd, iErrorLevel = OK_ERROR, iCurrentError = OK_ERROR;
     NTFS_VOL_INFO nvVol;
     PNTFS_MFT_ENTRY pEntry = NULL;
     uint64_t ulNum;
     int bCompact, bErasedOnly, bMd5, bSingle;
     FILE *fOutput;
     char *pOutputFilename = NULL, *pErrorFilename = NULL,
          *pDebugFilename = NULL;
#ifdef _WIN32
     char *pUserFilename = NULL;
#endif
     DT_LOCALE_TYPE mylocale = DT_SET_LOCALE;
     static struct option long_opts[] = {
          {"help", 0, 0, 'h'},
          {"verbose", 0, 0, 'v'},
          {"erased-only", 0, 0, 'e'},
          {"no-sequence-check", 0, 0, 'n'},
          {"md5", 0, 0, 'm'},
          {"Dump", 0, 0, 'D'},
	  {"utf8", 0, 0, '8'},
#ifndef _WIN32
          {"skip", 1, 0, 's'},
#endif
          {"index", 1, 0, 'i'},
#ifdef _WIN32
          {"user", 1, 0, 'u'},
          {"debug", 1, 0, 'd'},
#else
          {"debug", 2, 0, 'd'},
#endif
          {"error-output", 1, 0, 'x'},
          {"output", 1, 0, 'o'},
          {0, 0, 0, 0}
     };
#ifdef _WIN32
#define OPTSTRING "hvenmD8i:u:d:x:o:"
#else
#define OPTSTRING "hvenmD8s:i:d::x:o:"
#endif

     FErrorOutput = stderr;
     FIoDebug = NULL;
     BIoDump = FALSE;
#ifdef _WIN32
     FUserFile = NULL;
#else
     nvVol.ulSkip = 0;
#endif /* _WIN32 */
     bCompact = TRUE;
     bErasedOnly = FALSE;
     bMd5 = FALSE;
     fOutput = stdout;
     bSingle = FALSE;
     ulNum = 0;

     dt_setlocale("");

     while ((iArg =
             getopt_long (argc, argv, OPTSTRING, long_opts,
                          &iOptInd)) != -1) {
          switch (iArg) {
          case 'h':
               DisplayHelp (argv[0]);
               _fcloseall ();
               return OK_ERROR;
               break;

          case 'v':
               bCompact = FALSE;
               break;

          case 'm':
               bMd5 = TRUE;
               break;

          case 'e':
               bErasedOnly = TRUE;
               break;

          case 'n':
               BDoNotCheckSequence = TRUE;
               break;

	  case '8':
	       mylocale = DT_SET_UTF8;
	       break;

          case 'i':
               bSingle = TRUE;
               if (optarg)
/*              ulNum = _ttoi64(POptArg); */
                    ulNum = integer_arg (optarg);
               else {
                    Alarm (TA_PROG_FATAL,
                           "%s : MFT index value not found", argv[0]);
                    _fcloseall ();
                    return FATAL_ERROR;
               }
               break;

#ifdef _WIN32
          case 'u':
               MANDATORYFILENAME (optarg, pUserFilename, "user");
	       break;
#else /* _WIN32 */
          case 's':
               if (optarg)
                    nvVol.ulSkip = integer_arg (optarg);
               else {
                    Alarm (TA_PROG_FATAL,
                           "%s : skip value not found", argv[0]);
                    _fcloseall ();
                    return FATAL_ERROR;
               }
               break;
#endif /* _WIN32 */

          case 'o':
               MANDATORYFILENAME (optarg, pOutputFilename, "output");
               break;

          case 'x':
               MANDATORYFILENAME (optarg, pErrorFilename, "error");
               break;

          case 'D':
               BIoDump = TRUE;
               break;

          case 'd':
               if (optarg)
                    pDebugFilename = optarg;
               else
                    FIoDebug = stderr;
               break;

          default:
               Usage (argv[0]);
               _fcloseall ();
               return FATAL_ERROR;
               break;
          }                     /* switch(iArg) */
     }                          /* while((iArg = getopt(argc, argv,... */

     if (bSingle)
          bErasedOnly = FALSE;

     if (optind != (argc - 1)) {        /* 1 non-option argument */
          Usage (argv[0]);
          Alarm (TA_PROG_FATAL, "%s : wrong number of arguments", argv[0]);
          _fcloseall ();
          return FATAL_ERROR;
     }
#ifdef _WIN32
     if (argv[optind][0] >= TEXT ('a') && argv[optind][0] <= TEXT ('z')) {
          nvVol.tDrive = argv[optind][0] - TEXT ('a') + TEXT ('A');
     } else if (argv[optind][0] >= TEXT ('A') &&
                argv[optind][0] <= TEXT ('Z')) {
          nvVol.tDrive = argv[optind][0];
     } else {
          Alarm (TA_PROG_FATAL,
                 "%s : not a valid drive %c", argv[0], argv[optind][0]);
          _fcloseall ();
          return FATAL_ERROR;
     }
#else /* _WIN32 */
     nvVol.file = argv[optind];
#endif /* _WIN32 */

     OUTPUTFILE (pOutputFilename, fOutput, "output");

#ifdef _WIN32
     OUTPUTFILE (pUserFilename, FUserFile, "user");
#endif /* _WIN32 */

     OUTPUTFILE (pDebugFilename, FIoDebug, "debug");
     if (BIoDump && FIoDebug == NULL)
          FIoDebug = stderr;
     FInfoDebug = FIoDebug;
#ifdef _WIN32
     FSecurDebug = FIoDebug;
#endif /* _WIN32 */

     OUTPUTFILE (pErrorFilename, FErrorOutput, "error");

     set_wc_output_function(mylocale);

#ifdef _WIN32
     if ((!bCompact) || FUserFile)
          InitSecurity ();
#endif /* _WIN32 */

     DumpSystemInfo (fOutput, "dirmft", "", bCompact);
     fprintf (fOutput, "\n");

     iErrorLevel = OpenVolume (&nvVol);

     if (iErrorLevel == OK_ERROR) {
          if (bCompact)
               fprintf (fOutput, COMPACT_HEADER);

          iErrorLevel = InitEntry (&nvVol, &pEntry, ulNum, bErasedOnly);
          if ((iErrorLevel == OK_ERROR) &&
              ((pEntry->wRefCount == 0) || (!bErasedOnly)))
               iErrorLevel = DumpFileInfo (fOutput, pEntry, bMd5, bCompact);

          ulNum++;
          while (((ulNum * nvVol.dwMftRecordSize) <
                  nvVol.pMftData->ulStreamSize)
                 && (iErrorLevel != FATAL_ERROR) && (!bSingle)) {
               iErrorLevel = CleanEntry (pEntry);
               if (iErrorLevel == FATAL_ERROR)
                    break;
               iErrorLevel = ReadEntry (pEntry, ulNum, bErasedOnly);
               ulNum++;
               if ((iErrorLevel == OK_ERROR) &&
                   ((pEntry->wRefCount == 0) || (!bErasedOnly)))
                    iErrorLevel =
                         DumpFileInfo (fOutput, pEntry, bMd5, bCompact);
          }

     }
     if (pEntry)
          iCurrentError = FreeEntry (pEntry);

     iErrorLevel = max (iCurrentError, iErrorLevel);

     iCurrentError = CloseVolume (&nvVol);

     iErrorLevel = max (iErrorLevel, iCurrentError);

#ifdef _WIN32
     if ((!bCompact) || FUserFile) {
          iCurrentError = FreeSidHash ();
          iErrorLevel = max (iErrorLevel, iCurrentError);
     }
#endif /* _WIN32 */

     _fcloseall ();
     return iErrorLevel;
}
