/******************************************************************************
 *
 * dt_locale.c - Locale support code. Part of the DirTools project.
 *
 * $Revision: 1.2 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <locale.h>
#include <string.h>
#include <wchar.h>
#ifdef _WIN32
#include <windows.h>
#endif

#include "dt_error.h"
#include "dt_locale.h"
#include "macros.h"

DT_LOCALE_TYPE mylocale = DT_SET_LOCALE;

int dt_locale_fputwc(FILE *, const wchar_t);
int dt_utf8_fputwc(FILE *, const wchar_t);
int (*dt_fputwc)(FILE *, wchar_t) = dt_locale_fputwc;

char *
dt_getlocale (void)
{
	if (mylocale == DT_SET_LOCALE) {
		return setlocale(LC_ALL, NULL);
	} else if (mylocale == DT_SET_UTF8) {
		return ".UTF-8";
	} else {
		ALARM (TA_PROG_ERR, "Unable to find the locale the program use");
		return "";
	}
}

int
dt_setlocale (const char *locale)
{
	const char *result;

	if (locale == NULL) {
		result = setlocale(LC_ALL, "");
	} else {
		result =setlocale(LC_ALL, locale);
	}
	if (result == NULL) {
		ALARM(TA_PROG_WARN, "Failed to set locale, "
			"using default \"%s\"", dt_getlocale);
		return -1;
	}
	return 0;
}

void
dt_get_unicode (unsigned char *pBuffer, wchar_t *pText, uint32_t dwWcharNum)
{
     uint32_t dwCount;

     for (dwCount = 0; dwCount < dwWcharNum; dwCount++)
          pText[dwCount] = (wchar_t) DT_GETU16 (pBuffer + (2 * dwCount));

}

size_t
dt_mbstowcs(wchar_t *output, const char *input, size_t max)
{
	size_t result;
#ifdef HAVE_MBSINIT
	mbstate_t mbstate;
#endif

#ifdef HAVE_MBSINIT
	memset(&mbstate,0,sizeof(mbstate_t));
	result = mbsrtowcs(output, &input, max, &mbstate);
#else
	result = mbstowcs(output, input, max);
#endif

	if (result == (size_t)(-1)) {
		ALARM (TA_PROG_ERR, "Failed to convert multi-byte string"
			" to wide char string");
	} else {
		output[result] = L'\0';
	}

#ifdef HAVE_MBSINIT
	if (! mbsinit(&mbstate)) {
		ALARM (TA_PROG_ERR, "Failed to convert multi-byte string"
			" to wide char string");			
	}
#endif
	
        return result;
}

int
dt_locale_fputwc(FILE *stream, const wchar_t mywchar)
{
	int result;
	char output[MB_CUR_MAX + 1];
#ifdef HAVE_MBSINIT
	mbstate_t mbstate;
#endif

#ifdef HAVE_MBSINIT
	memset(&mbstate,0,sizeof(mbstate_t));
#else
	wctomb(NULL, 0);
#endif

#ifdef HAVE_MBSINIT
	result = wcrtomb(output, mywchar, &mbstate);
#else
	result = wctomb(output, mywchar);
#endif
	if (result == -1) {
		ALARM (TA_PROG_ERR, "Can't output 0x%x to current"
			" locale", (int)mywchar);
		return 0;
	}
#ifdef HAVE_MBSINIT
	if (! mbsinit(&mbstate)) {
		ALARM (TA_PROG_ERR, "Can't output 0x%x to current"
			" locale", (int)mywchar);
		return 0;			
	}
#endif

	output[result] = 0;
	fputs(output, stream);
	return result;
}

int
dt_utf8_fputwc(FILE *stream, const wchar_t mywchar)
{
	int value = (int) mywchar;
	
	if (value <= 0x7F) {
		fputc(value, stream);
		return 1;
	} else if (value <= 0x7FF) {
		fputc((int)(0xC0 | ((value & 0x7C0) >> 6)), stream);
		fputc((int)(0x80 | (value & 0x003F)), stream);
		return 2;
	} else if (value <= 0xFFFF) {
		fputc((int)(0xE0 | ((value & 0xF000) >> 12)), stream);
		fputc((int)(0x80 | ((value & 0xFC0) >> 6)), stream);
		fputc((int)(0x80 | (value & 0x003F)), stream);
		return 3;
	}
	ALARM (TA_PROG_ERR, "Can't output in UTF-8 value 0x%x",
		value);
	return 0;
}

void
set_wc_output_function(DT_LOCALE_TYPE choice)
{
	if (choice == DT_SET_UTF8) {
		dt_fputwc = dt_utf8_fputwc;
	} else if (choice == DT_SET_LOCALE) {
		dt_fputwc = dt_locale_fputwc;
	} else {
		ALARM (TA_PROG_FATAL, "Specified locale is unknown");
	}
	mylocale = choice;
}

int
dt_fputwcs(FILE *stream, const wchar_t *text)
{
	int index;
	int num = 0;

	for (index = 0; text[index] != L'\0'; index++) {
		num += dt_fputwc(stream, text[index]);
	}

	return num;
}
