/******************************************************************************
 *
 * output.c - NTFS file attributes dumping. Part of the DirTools project.
 *
 * $Revision: 1.2 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <inttypes.h>
#ifdef _WIN32
#include <windows.h>
#else /* _WIN32 */
#include "wincompat.h"
#endif

#include "output.h"
#include "dt_error.h"
#include "dt_locale.h"

/**********************************************************************
 * DumpHexBytes : print to file hex dump of a buffer
 *
 * - [in] fOutput : output file descriptor
 * - [in] pData : pointer to buffer
 * - [in] dwSize : buffer size in bytes
 *
 *********************************************************************/
void
DumpHexBytes (FILE * fOutput, unsigned char *pData, uint32_t dwSize)
{
     uint32_t dwOffset = 0, dwPos;
     uint16_t wLoop;

     while (dwOffset < dwSize) {
          fprintf (fOutput, "%04" PRIx32 ": ", dwOffset);
          for (wLoop = 0; wLoop < 16; wLoop++) {
               dwPos = wLoop + dwOffset;
               if (dwPos < dwSize)
                    fprintf (fOutput, "%02x ", pData[dwPos]);
               else
                    fprintf (fOutput, "   ");

               if (wLoop == 7)
                    fprintf (fOutput, " ");
          }                     /*  for (wLoop = 0; wLoop < 16; wLoop ++) */

          fprintf (fOutput, "  ");

          for (wLoop = 0; wLoop < 16; wLoop++) {
               dwPos = wLoop + dwOffset;
               if (dwPos < dwSize) {
                    if (isprint (pData[dwPos])) {
			 dt_fputwc(fOutput, (wchar_t)pData[dwPos]);
		    } else
                         fputs (".", fOutput);
               }

          }                     /*  for (wLoop = 0; wLoop < 16; wLoop ++) */

          fprintf (fOutput, "\n");
          dwOffset += 16;
     }                          /* while (dwOffset < dwSize) */
}

/**********************************************************************
 * DumpRawBytes : write to file the content of a buffer
 *
 * - [in] fOutput : output file descriptor
 * - [in] pData : pointer to buffer
 * - [in] dwSize : buffer size in bytes
 *
 *********************************************************************/
void
DumpRawBytes (FILE * fOutput, unsigned char *pData, uint32_t dwSize)
{
     fwrite (pData, sizeof (unsigned char), dwSize, fOutput);
}

/**********************************************************************
 * DumpAttribute : print to file information from attribute structure
 *
 * - [in] fOutput : output file descriptor
 * - [in] pAttr : pointer to the attribute structure
 *
 *********************************************************************/
void
DumpAttribute (FILE * fOutput, PNTFS_ATTRIBUTE pAttr)
{
     uint64_t ulCount;

     fprintf (fOutput, "   Type : 0x%" PRIx32 "\n", pAttr->dwType);
     fprintf (fOutput,
              (pAttr->bResident ? "   Resident\n" : "   Non-resident\n"));

     if (pAttr->wNameLength) {
	     fputs("   Name: ", fOutput);
	     dt_fputwcs(fOutput, pAttr->wcName);
	     fputs("\n", fOutput);
     }

     fprintf (fOutput, "   Stream size : 0x%" PRIx64 "\n",
              pAttr->ulStreamSize);
     if (pAttr->wCompressed)
          fprintf (fOutput, "   Compressed stream (Engine : 0x%" PRIx16 ")\n",
                   pAttr->wCompEngine);

     if (!pAttr->bResident) {
          fprintf (fOutput,
                   "   # of runs : %" PRIu64 "\n", pAttr->ulRunlistSize);
          fprintf (fOutput,
                   "   # of clusters : 0x%" PRIx64 "\n", pAttr->ulVcnsNumber);
     }

     for (ulCount = 0; ulCount < pAttr->ulRunlistSize; ulCount++) {
          fprintf (fOutput, "   Run #%" PRIu64 "\n", ulCount);
          fprintf (fOutput,
                   "    Lcn 0x%" PRIx64 ", length 0x%" PRIx64 "\n",
                   pAttr->pRunlist[ulCount].ulLcn,
                   pAttr->pRunlist[ulCount].ulLength);
     }

}

/**********************************************************************
 * DumpEntry : print to file information from entry structure
 *
 * - [in] fOutput : output file descriptor
 * - [in] pEntry : pointer to the entry structure
 *
 *********************************************************************/
void
DumpEntry (FILE * fOutput, PNTFS_MFT_ENTRY pEntry)
{
     uint16_t wCount;

     fprintf (fOutput,
              "Base record number : %" PRIu64 "\n", pEntry->ulNumber);
     fprintf (fOutput, " Number of hard links : %" PRIu16 "\n",
              pEntry->wRefCount);
     fprintf (fOutput, " Number of attributes : %" PRIu16 "\n",
              pEntry->wAttrCount);

     for (wCount = 0; wCount < pEntry->wAttrCount; wCount++) {
          fprintf (fOutput, "  Attribute #%" PRIu16 "\n", wCount);
          DumpAttribute (fOutput, &pEntry->pAttr[wCount]);
     }
     fprintf (fOutput, "\n");
}

/**********************************************************************
 * ProcessRawBytes : generic version of DumpRawBytes used by
 *                   ProcessStream (diskio.c)
 *
 * - [in] pArg : should a file descriptor pointer
 * - [in] pBuffer : pointer to buffer
 * - [in] dwSize : buffer size in bytes
 *
 *********************************************************************/
void
ProcessRawBytes (void *pArg, unsigned char *pBuffer, uint32_t dwSize)
{
     DumpRawBytes ((FILE *) pArg, pBuffer, dwSize);
}

/**********************************************************************
 * ProcessHexBytes : generic version of DumpHexBytes used by
 *                   ProcessStream (diskio.c)
 *
 * - [in] pArg : should a file descriptor pointer
 * - [in] pBuffer : pointer to buffer
 * - [in] dwSize : buffer size in bytes
 *
 *********************************************************************/
void
ProcessHexBytes (void *pArg, unsigned char *pBuffer, uint32_t dwSize)
{
     DumpHexBytes ((FILE *) pArg, pBuffer, dwSize);
}
