/******************************************************************************
 *
 * winsid.c - convert SID to full name. Part of the DirTools project.
 *
 * $Revision: 1.1.1.1 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#ifdef _WIN32

#include <windows.h>

#include "dt_error.h"
#include "winsid.h"
#include "dt_memory.h"
#include "sysheader.h"

#define SID_HASH_SIZE 300
PTYPE_SID_INFO PSidHash[SID_HASH_SIZE];
#define SID_TYPE_NUMBER 10
LPTSTR PTypeSidList[SID_TYPE_NUMBER];

FILE *FUserFile = NULL;
FILE *FSecurDebug = NULL;

/**********************************************************************
 * InitSecurity : initialize the SID hash table and the SID type 
 *                definition
 *
 *********************************************************************/
void
InitSecurity ()
{
     uint16_t wCount;

     for (wCount = 0; wCount < SID_HASH_SIZE; wCount++) {
          PSidHash[wCount] = NULL;
     }

     PTypeSidList[SidTypeUser] = "User";
     PTypeSidList[SidTypeGroup] = "Group";
     PTypeSidList[SidTypeDomain] = "Domain";
     PTypeSidList[SidTypeAlias] = "Alias";
     PTypeSidList[SidTypeWellKnownGroup] = "Well known group";
     PTypeSidList[SidTypeDeletedAccount] = "Deleted account";
     PTypeSidList[SidTypeInvalid] = "Invalid";
     PTypeSidList[SidTypeUnknown] = "Unknown";
/* PTypeSidList[SidTypeComputer]="Computer"; */

}

/**********************************************************************
 * GetSidHash : looks for the textual version of a SID in the hash
 *              table, add it and its description if not found, gives
 *              the type and a description
 *
 * - [in] pSid
 * - [in] pTestSid
 * - [out] pComment
 * - [out] pType
 *
 * returns OK_ERROR in case of success, FATAL_ERROR in case of memory
 *         leak, SIMPLE_ERROR otherwise
 *
 *********************************************************************/
int
GetSidHash (PSID pSid, const char *pTestSid, char **pComment, char **pType)
{
     int iErrorLevel = OK_ERROR;
     PTYPE_SID_INFO pSidSearch;
     int bFound = FALSE;
     uint32_t dwNameSize = 0;
     uint32_t dwDomainSize = 0;
     char *pName = NULL;
     SID_NAME_USE TypeSid = SidTypeUnknown;
     uint16_t wHash;
     unsigned long ulH = 0, ulG;
     const char *pHashName = pTestSid;

     /* ElfHash algorithm used in UNIX ELF format
        IS IT COMPATIBLE WITH UNICODE STRINGS ???? */
     while (*pHashName) {
          ulH = (ulH << 4) + *pHashName++;
          if ((ulG = ulH & 0xF0000000))
               ulH ^= ulG >> 24;
          ulH &= ~ulG;
     }
     wHash = ulH % SID_HASH_SIZE;

     pSidSearch = PSidHash[wHash];
     *pComment = NULL;
     *pType = NULL;

     while (pSidSearch) {
          if (!lstrcmp (pTestSid, pSidSearch->pTextualSid)) {
               *pComment = pSidSearch->pCommentSid;
               *pType = pSidSearch->pTypeSid;
               bFound = TRUE;
               if (FSecurDebug)
                    fprintf (FSecurDebug,
                             "GetSidHash : SID %s found in hash\n", pTestSid);
               break;
          }
          pSidSearch = pSidSearch->pNext;
     }

     if (!bFound) {
          pSidSearch = DT_CALLOC (TYPE_SID_INFO, 1, "pSidSearch");

          pSidSearch->pTextualSid = NULL;
          pSidSearch->pNext = PSidHash[wHash];
          PSidHash[wHash] = pSidSearch;
          pSidSearch->pTextualSid = DT_CALLOC (char, lstrlen (pTestSid) + 1,
                                               "pTextualSid");
          lstrcpyn (pSidSearch->pTextualSid, pTestSid,
                    lstrlen (pTestSid) + 1);

          pSidSearch->pCommentSid = NULL;
          /* Get details about SID
             First call to get required buffer sizes */
          LookupAccountSid (NULL,       /* Query local computer */
                            pSid,       /* pointer to SID */
                            pSidSearch->pCommentSid,    /* p. name buffer */
                            &dwNameSize,        /* Name buffer size */
                            pSidSearch->pCommentSid,    /* p. domain buffer */
                            &dwDomainSize,      /* Domain buffer size */
                            &TypeSid);

          if (GetLastError () != ERROR_INSUFFICIENT_BUFFER) {
               ALARM (TA_SYS_ERR, "1st LookupAccountSid");
               iErrorLevel = SIMPLE_ERROR;
          } else {              /* GetLastError() == ERROR_INSUFFICIENT... */

               pSidSearch->pCommentSid = DT_CALLOC (unsigned char,
                                                    dwNameSize +
                                                    dwDomainSize + 1,
                                                    "pCommentSid");

               pName = DT_CALLOC (unsigned char, dwNameSize, "pName");


               if (!LookupAccountSid (NULL,     /* Query to local computer */
                                      pSid,     /* pointer to SID */
                                      pName,    /* pointer to name buffer */
                                      &dwNameSize,      /* Buffer size */
                                      pSidSearch->pCommentSid,  /* d. buffer */
                                      &dwDomainSize,    /* Buffer size */
                                      &TypeSid)) {
                    ALARM (TA_SYS_ERR, "2nd LookupAccountSid");
                    iErrorLevel = SIMPLE_ERROR;
               } else {         /* if (!LookupAccountSid(... */

                    pSidSearch->pTypeSid = PTypeSidList[TypeSid];
                    sprintf (pSidSearch->pCommentSid +
                             lstrlen (pSidSearch->pCommentSid), "\\");
                    sprintf (pSidSearch->pCommentSid +
                             lstrlen (pSidSearch->pCommentSid), "%s", pName);
                    if (FSecurDebug)
                         fprintf (FSecurDebug,
                                  "GeSidHash : add %s=%s\n",
                                  pSidSearch->pTextualSid,
                                  pSidSearch->pCommentSid);

                    *pComment = pSidSearch->pCommentSid;
                    *pType = pSidSearch->pTypeSid;

               }                /* if (!LookupAccountSid(... */

               DT_FREE (pName, "pName");

          }                     /* GetLastError() == ERROR_INSUFFICIENT_... */


     }
     /*  if (!bFound) */
     return iErrorLevel;
}

/**********************************************************************
 * FreeSidHash : frees the SID hash table, dump the content to file
 *               if FUserFile is not NULL
 *
 * returns OK_ERROR in case of success, FATAL_ERROR in case of memory
 *         leak
 *
 *********************************************************************/
int
FreeSidHash ()
{
     int iErrorLevel = OK_ERROR;
     uint16_t wCount;
     PTYPE_SID_INFO pSidSearch, pSidMemo;

     if (FUserFile) {
          DumpSystemInfo (FUserFile, "users", "", TRUE);
          fprintf (FUserFile, "\nSID|Type|Name\n");
     }

     for (wCount = 0; wCount < SID_HASH_SIZE; wCount++) {
          pSidSearch = PSidHash[wCount];
          while (pSidSearch != NULL) {
               pSidMemo = pSidSearch->pNext;

               if (pSidSearch->pTextualSid) {
                    if (FUserFile)
                         fprintf (FUserFile, "%s", pSidSearch->pTextualSid);

                    DT_FREE (pSidSearch->pTextualSid, "pTextualSid");
               }
               /*   if (pSidSearch->pTextualSid) */
               if (FUserFile)
                    fprintf (FUserFile, "|%s", pSidSearch->pTypeSid);

               if (pSidSearch->pCommentSid) {
                    if (FUserFile)
                         fprintf (FUserFile, "|%s", pSidSearch->pCommentSid);
                    DT_FREE (pSidSearch->pCommentSid, "pCommentSid");
               }
               /* if (pSidSearch->pCommentSid) */
               DT_FREE (pSidSearch, "pSidSearch");

               if (FUserFile)
                    fprintf (FUserFile, "\n");

               pSidSearch = pSidMemo;

          }                     /* while (pSidSearch != NULL) */
     }                          /* for (wCount = 0; wCount < SID_HASH_... */

     return iErrorLevel;
}
#endif /* _WIN32 */
